'use strict';

// This module is building out a Prosemirror object on the window element.
// This object is having all the necessary classes and methods to create
// a prosemirror editor instance.
//
// It is then mixing basic editor helpers with all the "mention" module helpers

import {Schema} from '../../../../node_modules/prosemirror-model';
import {EditorView} from '../../../../node_modules/prosemirror-view';
import {EditorState} from '../../../../node_modules/prosemirror-state';
import {MarkdownParser} from '../../../../node_modules/prosemirror-markdown';
import {menuBar, MenuItem} from '../../../../node_modules/prosemirror-menu';

import {Subscription} from '../../../../node_modules/subscription';

const markdownit = (require('../../../../node_modules/markdown-it'))('commonmark', {html: false})
                    .use(require('../../../../node_modules/markdown-it-flowdock'))
                    .disable(['hashtag', 'flowdock-url']);

import {basicSetupPlugin, basicSchemaSpec, buildMenuItems,
  basicMarkdownSerializer, basicMarkdownParserTokens} from './basic-setup';
import {mentionSpec, mentionMarkdownParserToken,
  mentionMarkdownSerializer, mentionAutocompletePlugins} from './mentions';
import { getPastedImage } from './basic-setup/plugins';


// method helper: like angular.extend

function extend() {
  var out = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
  var args = [];
  Array.prototype.push.apply(args, arguments);
  args.slice(1).forEach(function (obj) {
    if (obj) {
      for (var key in obj) {
        if (({}).hasOwnProperty.call(obj, key)) {
          out[key] = obj[key];
        }
      }
    }
  });
  return out;
}


// Get editor's schema based on editor's type.
// For now, it helps adding "mentions" to basic schema

function getSchema(type) {
  var nodes = basicSchemaSpec.nodes;
  if (type === 'comment' || type === 'content') {
    nodes = basicSchemaSpec.nodes.addBefore('text', 'mention', mentionSpec);
  }
  return new Schema({
    nodes: nodes,
    marks: basicSchemaSpec.marks
  });
}


// Get editor's plugins based on editor's type.
// Provide options to basic plugins helper method.
//
// Add mention autocomplete plugins if necessary

function getPlugins(options, type) {
  let plugins = basicSetupPlugin(options);
  if (type === 'comment' || type === 'content') {
    plugins = mentionAutocompletePlugins().concat(plugins);
  }
  return plugins;
}


function getMarkdownParserTokens(type) {
  if (type === 'comment' || type === 'content') {
    return extend({}, basicMarkdownParserTokens, {
      mention: mentionMarkdownParserToken
    });
  } else {
    return basicMarkdownParserTokens;
  }
}


function getMarkdownParser(schema, type) {
  return new MarkdownParser(schema, markdownit, getMarkdownParserTokens(type));
}


function getMarkdownSerializer(type) {
  let markdownSerializer = Object.assign(Object.create(basicMarkdownSerializer), basicMarkdownSerializer);
  if (type === 'comment' || type === 'content') {
    markdownSerializer.nodes.mention = mentionMarkdownSerializer;
  }

  return markdownSerializer;
}

window.ProseMirror = {
  EditorState: EditorState,
  EditorView: EditorView,
  getSchema: getSchema,
  getPlugins: getPlugins,
  getMarkdownParser: getMarkdownParser,
  getMarkdownSerializer: getMarkdownSerializer,
  Subscription: Subscription,
  buildMenuItems: buildMenuItems,
  menuBar: menuBar,
  MenuItem: MenuItem,
  getPastedImage: getPastedImage
};