'use strict';

var BaseStoreSearch = function() {
  this.query = '';
  this.listeners = [];
};

BaseStoreSearch.prototype.on = function(listener, immediate) {
  this.listeners.push(listener);
  if (immediate === undefined || immediate) {
    listener.call(this);
  }
  var self = this;
  return function() {
    delete self.listeners[self.listeners.indexOf(listener)];
  };
};

BaseStoreSearch.prototype.$on = function(scope, listener, immediate) {
  var self = this;
  scope.$on('$destroy', this.on(function() {
    scope.$applyAsync(function() {
      listener.call(self);
    });
  }, immediate));
};

BaseStoreSearch.prototype.trigger = function() {
  this.listeners.forEach(function(listener) {
    listener.call(this);
  }, this);
};

BaseStoreSearch.prototype.set = function(value) {
  this.query = value;
  this.trigger();
};

BaseStoreSearch.prototype.reset = function(clearListeners) {
  this.query = '';
  this.trigger();
  if (clearListeners) {
    this.listeners = [];
  }
  return this;
};

BaseStoreSearch.prototype.filter = function(objects) {
  return objects.filter(this.match.bind(this));
};

BaseStoreSearch.prototype.match = function(obj) {
  throw new Exception('Not implemented');
};


var StoreSearch = function(options, lunr) {
  BaseStoreSearch.call(this);

  this.options = angular.extend({}, StoreSearch.defaultOptions, options);
  this.lunr = lunr;
  this.lunrCache = null;
  this.current = null;
  this.cache = {};
  this.matchersCache = {};

  if (lunr && typeof(this.options.matchers[this.options.textMatcher]) === 'undefined') {
    this.options.matchers._text = function(obj, value) {
      if (this.lunrCache === null) {
        this.lunrCache = this.lunr.search(value);
      }
      for (var i = 0; i < this.lunrCache.length; i++) {
        if (this.lunrCache[i].ref == obj.id) {
          return true;
        }
      }
      return false;
    };
  }
};

StoreSearch.prototype = new BaseStoreSearch();

StoreSearch.defaultOptions = {
  autoTags: false,
  ignoreTags: [],
  discardTags: [],
  ignoreTextChars: [],
  extractors: [],
  sanitizers: [],
  matchers: {},
  textMatcher: '_text'
};

StoreSearch.extractRegexpFromString = function(re, value, multiple, mapper) {
  if (typeof(multiple) === 'function') {
    mapper = multiple;
    multiple = true;
  }
  var res;
  var results = [];
  var out = value;
  while ((res = re.exec(value)) !== null) {
    var v = res[1].toLowerCase();
    if (v[0] === '"') {
      v = v.substr(1, v.length - 2);
    }
    results.push(v)
    out = out.replace(res[0], '');
  }
  if (mapper) {
    results = results.map(mapper).filter(function(v) { return v !== null; });
  }
  if (results.length === 0) {
    results = null;
  } else if (multiple !== undefined && !multiple) {
    results = results[results.length - 1];
  }
  return {'results': results, 'out': out};
};

StoreSearch.extractFlagFromString = function(flag, value) {
  return {'present': value.indexOf(flag) !== -1, 'out': value.replace(flag, '')};
};

StoreSearch.extraxtRegexpFiltersFromString = function(re, name, value, filters, multiple, mapper) {
  var r = StoreSearch.extractRegexpFromString(re, value, multiple, mapper);
  if (r.results) {
    filters[name] = r.results;
  }
  return r.out;
};

StoreSearch.extractFlagFiltersFromString = function(flag, name, value, filters) {
  var r = StoreSearch.extractFlagFromString(flag, value);
  if (r.present) {
    filters[name] = true;
  }
  return r.out;
};

StoreSearch.tagRegexp = function(name) {
  return new RegExp(name + ":([^ ]+)", 'gi');
};

StoreSearch.extractTagsFromString = function(value, ignoreTags, discardTags) {
  var re = StoreSearch.tagRegexp('([a-z]+)');
  var res;
  var results = {};
  var out = value;
  while ((res = re.exec(value)) !== null) {
    var tag = res[1].toLowerCase();
    if (ignoreTags && ignoreTags.indexOf(tag) !== -1) {
      continue;
    }
    if (!discardTags || discardTags.indexOf(tag) === -1) {
      if (typeof(results[tag]) === 'undefined') {
        results[tag] = [];
      }
      results[tag].push(res[2].toLowerCase());
    }
    out = out.replace(res[0], '');
  }
  return {'results': results, 'out': out};
};

StoreSearch.prototype.set = function(value) {
  if (!value || $.trim(value) === '') {
    this.reset();
    return;
  }

  this.query = value;
  var self = this;
  var filters = {};

  for (var i = 0; i < this.options.extractors.length; i++) {
    value = this.options.extractors[i].apply(this, [value, filters]);
  }

  if (this.options.autoTags) {
    var r = StoreSearch.extractTagsFromString(value, this.options.ignoreTags, this.options.discardTags);
    angular.extend(filters, r.results);
    value = r.out;
  }

  for (i = 0; i < this.options.ignoreTextChars.length; i++) {
    value = value.replace(this.options.ignoreTextChars[i], '');
  }
  value = $.trim(value);
  if (value !== '' && this.options.textMatcher) {
    filters[this.options.textMatcher] = value;
    this.lunrCache = null;
  }

  for (i = 0; i < this.options.sanitizers.length; i++) {
    this.options.sanitizers[i].apply(this, [filters]);
  }

  if (this.current) {
    Object.keys(this.current).forEach(function(name) {
      if (typeof(this.matchersCache[name]) === 'undefined') {
        return;
      }
      if (typeof(filters[name]) === 'undefined' || !angular.equals(this.current[name], filters[name])) {
        delete this.matchersCache[name];
      }
    }, this);
  }
  this.current = filters;
  this.cache = {};
  this.trigger();
  return this;
};

StoreSearch.prototype.clearCache = function() {
  this.cache = {};
  this.matchersCache = {};
  this.lunrCache = null;
};

StoreSearch.prototype.clearObjectCache = function(obj) {
  if (typeof(this.cache[obj.id]) !== 'undefined') {
    delete this.cache[obj.id];
  }
  if (this.current) {
    Object.keys(this.current).forEach(function(name) {
      if (typeof(this.matchersCache[name]) !== 'undefined' && typeof(this.matchersCache[name][obj.id]) !== 'undefined') {
        delete this.matchersCache[name][obj.id];
      }
    }, this);
  }
};

StoreSearch.prototype.clearCacheBindToStore = function(store) {
  var self = this;
  return store._offMany([
    store.on('inserted', function() {
      self.lunrCache = null;
    }),
    store.on('updated', function(obj) {
      self.lunrCache = null;
      self.clearObjectCache(obj);
    })
  ]);
};

StoreSearch.prototype.reset = function(clearListeners) {
  this.current = null;
  this.clearCache();
  return BaseStoreSearch.prototype.reset.call(this, clearListeners);
};

StoreSearch.prototype.match = function(obj) {
  if (!this.current) {
    return true;
  }
  if (typeof(this.cache[obj.id]) !== 'undefined') {
    return this.cache[obj.id];
  }
  var matches = true;
  angular.forEach(this.current, function(value, name) {
    if (!matches) {
      return;
    }
    var matcher;
    if (typeof(this.options.matchers[name]) === 'undefined') {
      if (typeof(this.options.matchers._fallback) === 'undefined') {
        matches = false;
        return;
      }
      matcher = this.options.matchers._fallback;
    } else {
      matcher = this.options.matchers[name];
    }
    if (typeof(this.matchersCache[name]) === 'undefined') {
      this.matchersCache[name] = {};
    } else if (typeof(this.matchersCache[name][obj.id]) !== 'undefined') {
      matches = this.matchersCache[name][obj.id];
      return;
    }
    matches = matcher.apply(this, [obj, value, name]);
    this.matchersCache[name][obj.id] = matches;
  }, this);
  this.cache[obj.id] = matches;
  return matches;
};

var StoreLunrMixin = {
  insert: function(obj) {
    obj = Store.prototype.insert.apply(this, arguments);
    this.lunr.add(obj);
    return obj;
  },
  update: function(obj) {
    obj = Store.prototype.update.apply(this, arguments);
    this.lunr.update(obj);
    return obj;
  },
  updateById: function(id) {
    this.lunr.remove({id: id});
    return Store.prototype.updateById.apply(this, arguments);
  },
  remove: function(obj) {
    obj = Store.prototype.remove.apply(this, arguments);
    this.lunr.remove(obj);
    return obj;
  },
  _clear: function() {
    Store.prototype._clear.apply(this, arguments);
    this.lunr = this._createLunrIndex();
  },
  _createLunrIndex: function() {
    return lunr(function() {
      this.field('title');
      this.ref('id');
    });
  }
};

var StoreSearchMixin = {
  searchOptions: {},
  createSearch: function(query, clearCacheBindToStore) {
    var s = new StoreSearch(this.searchOptions, this.getLunrIndex());
    if (clearCacheBindToStore) {
      s.clearCacheBindToStore(this);
    }
    if (query) {
      s.set(query);
    }
    return s;
  },
  search: function(query) {
    return this.createSearch(query).filter(this.data);
  },
  getLunrIndex: function() {
    return this.lunr;
  }
};


var RemoteStoreSearch = function(queryExecuter) {
  BaseStoreSearch.call(this);
  this.results = null;
  this.executeQuery = queryExecuter;
  this.runningSearchs = 0;
};

RemoteStoreSearch.prototype = new BaseStoreSearch();

RemoteStoreSearch.prototype.reset = function(clearListeners) {
  this.results = null;
  return BaseStoreSearch.prototype.reset.call(this, clearListeners);
};

RemoteStoreSearch.prototype.set = function(value) {
  var self = this;
  self.query = value;
  self.runningSearchs += 1;
  return this.executeQuery(value).then(function(results) {
    self.results = results;
    self.trigger();
    self.runningSearchs -= 1;
  }, function(reason) {
    if (reason !== 'abort') {
      self.results = [];
      self.trigger();
    }
    self.runningSearchs -= 1;
  });
};

RemoteStoreSearch.prototype.isRunning = function() {
  return this.runningSearchs > 0;
};

RemoteStoreSearch.prototype.match = function(obj) {
  return this.results === null || this.results.indexOf(obj.id) !== -1;
};